import numpy as np
from scipy import integrate
import plasmapy.formulary as pp_form
import astropy.units as units
from .constants import *
import warnings

#!Maybe not the best idea, but fine for now...
warnings.filterwarnings("ignore")


class DaligaultGamma:
    """
    class DaligaultGamma generates all necessary parameters and outputs for analysis
    """

    def __init__(
        self, E, Te, n, theta_cutoff=1, q_kF_list=np.linspace(1e-3, 10, 1000)
    ) -> None:
        """Function returns all parameters

        Args:
            E (array): Energy array in eV
            Te (float): electronic temperature in eV
            theta (float): degeneracy parameter
            n (float): carrier density of material (1/m^3)
            theta_cutoff (float): Te (eV) cutoff in which we go from quantum to classical chemical potential
            q_kF_list (array): array of q/kF points used for v(q) and G(q/kF), defaults to np.linspace(1e-3, 10, 1000) 1/m
        """
        self.E = E
        self.Te = Te
        self.theta = pp_form.quantum_theta(Te * units.eV, n).value
        self.n = n
        self.theta_cutoff = theta_cutoff
        self.q_kF_list = q_kF_list

    #!---------------------------------------------------------------
    # * Start of density- and theta-dependent parameter section

    def get_kF(n):
        return (3 * np.pi**2 * n) ** (1 / 3)

    def get_T_from_theta(n, theta):
        Ef = pp_form.quantum.Ef_(n * units.m**-3).value
        return theta * Ef / constants.kB

    def get_mu_quantum(n, theta):
        T = get_T_from_theta(n, theta)
        Ef = pp_form.quantum.Ef_(n * units.m**-3).value
        return Ef * (1 - 1 / 3 * ((np.pi * constants.kB * T) / (2 * Ef)) ** 2)

    # Eq. 5.26 on pp. 40 in Fetter and Walecka
    def get_mu_classical(n, theta):
        T = get_T_from_theta(n, theta)
        mu = (
            constants.kB
            * T
            * np.log(n / 2 * pp_form.quantum.lambdaDB_th_(T * units.K).value ** 3)
        )
        return mu

    def get_mu(n, theta, theta_cutoff=1):
        if theta < theta_cutoff:
            mu = get_mu_quantum(n, theta)
        elif theta >= theta_cutoff:
            mu = get_mu_classical(n, theta)
        return mu

    def get_omega_p(n):
        return pp_form.plasma_frequency(n * units.m**-3, particle="e-").value

    # * End of density-dependent parameter section
    #!---------------------------------------------------------------

    #!---------------------------------------------------------------
    # * Start of Eq. (4) section

    def get_u_integral(u, k_kF, q_kF, theta, sq_mu):
        integral_first_term = (1 / (1 - np.exp(((2 * q_kF) / theta) * u))) * (
            1
            / (
                1
                + np.exp((-(2 * q_kF) / theta) * u)
                * np.exp((1 / theta) * (sq_mu - k_kF**2))
            )
        )
        integral_second_term = np.log(
            (1 + np.exp((1 / theta) * (sq_mu - (u + q_kF / 2) ** 2)))
            / (1 + np.exp((1 / theta) * (sq_mu - (-u + q_kF / 2) ** 2)))
        )

        return integral_first_term * integral_second_term

    def get_G(k_kF, q_kF, theta, sq_mu, F_12, F_n12):

        integral, error = integrate.quad(
            get_u_integral,
            q_kF / 2 - k_kF,
            q_kF / 2 + k_kF,
            args=(k_kF, q_kF, theta, sq_mu),
        )

        G = (
            (9 * np.pi / (16 * np.sqrt(2)))
            * (theta ** (5 / 2))
            * (F_12 / F_n12) ** (3 / 2)
            * (q_kF / k_kF)
            * integral
        )

        return G, error

    # * End of Eq. (4) section
    #!---------------------------------------------------------------

    #!---------------------------------------------------------------
    # * Start of v(q) section

    def get_k_sc(n, theta, F_12, F_n12):
        T = get_T_from_theta(n, theta)
        k_DH = (
            1 / pp_form.lengths.Debye_length(T * units.K, n * units.m**-3).value
        )  # 1/m
        k_sc = k_DH * np.sqrt(F_n12 / F_12)  # 1/m

        return k_sc

    def get_v(q, k_sc):
        v = 1 / (q**2 + k_sc**2)
        return v

    # * End of v(q) section
    #!---------------------------------------------------------------

    #!---------------------------------------------------------------
    # * Start of Eq. (3) section

    def get_q_integrand(q, q_kF, k_kF, k_sc, theta, sq_mu, F_12, F_n12):
        v2 = np.abs(get_v(q, k_sc)) ** 2
        G = get_G(k_kF, q_kF, theta, sq_mu, F_12, F_n12)[0]

        q_integrand = v2 * G
        return q_integrand

    def get_Eq_3(q_list, kF, k_kF, k_sc, theta, sq_mu, F_12, F_n12):

        q_integrand = np.array(
            [
                get_q_integrand(q, q / kF, k_kF, k_sc, theta, sq_mu, F_12, F_n12)
                for q in q_list
            ]
        )
        q_integral = integrate.cumtrapz(q_integrand, q_list, initial=0)[-1]

        return k_sc**3 * q_integral

    def get_gamma_omega_p(q_kF_list, k_kF_list, n, theta, theta_cutoff):
        print(
            f"n = {n:.2e} 1/m^3, theta = {theta:.2f}, Te={get_T_from_theta(n, theta)/constants.eV_to_K:.2f}"
        )
        # * Assign values independent of k and q
        kF = get_kF(n)
        mu = get_mu(n, theta, theta_cutoff)
        sq_mu = mu / pp_form.quantum.Ef_(n * units.m**-3).value
        F_12 = np.real(pp_form.mathematics.Fermi_integral(sq_mu / theta, 1 / 2))
        F_n12 = np.real(pp_form.mathematics.Fermi_integral(sq_mu / theta, -1 / 2))
        k_sc = get_k_sc(n, theta, F_12, F_n12)
        q_list = q_kF_list * get_kF(n)
        k_list = k_kF_list * get_kF(n)

        gamma_omega_p = np.array(
            [
                get_Eq_3(q_list, kF, k / kF, k_sc, theta, sq_mu, F_12, F_n12)
                for k in k_list
            ]
        )

        return gamma_omega_p

    #!---------------------------------------------------------------
    # * Start of Eq. (3) section

    def get_gamma(self):
        """
        Function calculates Daligault gamma
        Returns:
            E_gamma (2D array): Energy-mu (eV) and Daligault scattering rate (1/fs)
        """

        k_kF_list = (
            np.sqrt(2 * constants.m_e * self.E * constants.eV_to_J) / constants.hbar
        ) / get_kF(self.n)

        gamma_list = (
            np.array(
                get_gamma_omega_p(
                    self.q_kF_list, k_kF_list, self.n, self.theta, self.theta_cutoff
                )
            )
            * get_omega_p(self.n)
            * 1e-15
        )
        # 1/fs
        E_mu_list = (
            self.E - get_mu(self.n, self.theta, self.theta_cutoff) / constants.eV_to_J
        )

        E_gamma = np.array([E_mu_list, gamma_list])  # eV and 1/fs
        return E_gamma



import numpy as np
import math
from .constants import *
import warnings

warnings.filterwarnings("ignore")


class FGParams:
    """
    class FLParams generates Fermi gas parameters
    """

    def __init__(self, Te, N, V) -> None:
        """Function returns all parameters

        Args:
            Te (float): electronic temperature in eV
            N (int): Number of electrons in cell
            V (float): Volume of cell in Angstrom cubed
            Ef (float): Fermi energy of free electron gas in eV
            mu (float): Chemical potential of free electron gas in eV
            V_au (float): Volume in atomic units
            n_au (float): particle density (N/V) in a.u.
            kf (float): Fermi wave vector in a.u.
            rs (float): Wigner-Seitz radius in a.u.
            N0 (float): DOS at Ef in a.u.
        """
        self.Te = Te
        self.N = int(N)
        self.V = V
        self.Ef = (
            hbar_Js**2
            / (2 * m_e)
            * (3 * np.pi**2 * self.N / (self.V * 1e-30)) ** (2 / 3)
        ) * J_to_eV
        self.mu = self.Ef * (1 - 1 / 3 * ((np.pi * self.Te) / (2 * self.Ef)) ** 2)
        # Use atomic units:  hbar = m = e = 1  a.u
        self.V_au = self.V * (1.88973) ** 3  # a0^3
        self.n_au = self.N / self.V_au
        self.kf = (3 * np.pi**2 * self.n_au) ** (1 / 3)
        self.rs = (3 / (4 * np.pi * self.n_au)) ** (1 / 3)
        self.N0 = 3 * self.n_au / (self.kf**2)



class VaspParam:
    """
    class VaspParam reads and produces all desired parameters from a VASP calcualtion
    """

    def __init__(self, file) -> None:
        """Function reads through VASP output self.outcar (self.outcar) and parses out key parameters

        Args:
            file (string): the path to the OUTCAR file
        Returns:
            file (string): the path to the OUTCAR file
            outcar (list): the OUTCAR file written to a list
            algo (string): GW calculation algorithm
            finite_temp (string): GW calculation type (T/F)
            full_freq (string): GW calculation type (T/F)
            Te (float): Electronic temperature in eV
            Ef (float): Fermi energy of calculation
            volume (float): Volume of cell in A^3
            nelect (float): Total number of electrons
            ispin (int): Spin of system
            N (int): Number of electrons accounting for degeneracy
            num_ion_type (array): Number of each ion species
            pomass (array): Mass of each ion species
            mass (float): Total mass of the cell
            rho (float): Density of the cell in g/cc
            nbands (int): Number of bands in calculation
            nbandsgw (int): Number of bands included in GW calculation
            nkpts (int): Number of irreducible kpoints in calculation
            nkdim (int): Total number of kpoints in calculation
            nomega (int): Number of frequency grid points in GW calculation
            cshift (float): Shift of the Hilbert transformation for GW calculation
        """

        self.file = file
        self.outcar = [line for line in open(self.file, "r")]
        self.algo = [line for line in self.outcar if "ALGO" in line][0].split()[-1]
        self.finite_temp = [line for line in self.outcar if "LFINITE_T" in line][
            -1
        ].split()[-5]
        self.full_freq = [line for line in self.outcar if "LSELFENERGY" in line][
            -1
        ].split()[-7]
        # use Te instead of sigma due to sigma being used for self-energy elsewhere
        self.Te = float(
            [line for line in self.outcar if "SIGMA" in line][0].split()[-1]
        )
        self.Ef = float(
            [line for line in self.outcar if "E-fermi" in line][0].split()[-1]
        )
        self.volume = float(
            [line for line in self.outcar if "volume" in line][0].split()[-1]
        )
        self.nelect = float(
            [line for line in self.outcar if "NELECT" in line][0].split()[2]
        )
        self.ispin = int(
            [line for line in self.outcar if "ISPIN" in line][0].split()[2]
        )
        self.N = int(self.ispin * self.nelect / 2)
        self.num_ion_type = [line for line in self.outcar if "ions per type" in line][
            0
        ].split()[4:]
        self.pomass = [line for line in self.outcar if "POMASS" in line][-1].split()[2:]
        self.mass = sum(
            [
                float(ion) * float(mass)
                for ion, mass in list(zip(self.num_ion_type, self.pomass))
            ]
        )  # Dalton
        self.rho = (self.mass * 1.6605e24) / (self.volume * 1e24)  # g/cc
        self.nbands = int(
            [line for line in self.outcar if "NBANDS" in line][0].split()[-1]
        )
        self.nbandsgw = int(
            [line for line in self.outcar if "NBANDSGW " in line][-1].split()[-10]
        )
        self.nkpts = int(
            [line for line in self.outcar if "NKPTS" in line][0].split()[3]
        )
        self.nkdim = int(
            [line for line in self.outcar if "NKDIM" in line][0].split()[9]
        )
        self.nomega = int(
            [line for line in self.outcar if "NOMEGA" in line][0].split()[-1]
        )
        self.cshift = float(
            [line for line in self.outcar if "CSHIFT" in line][-1].split()[-9]
        )
